# -*- coding: utf-8 -*-


class Constants(object):
    """
    系统常量
    """
    # dict_load_data
    DICT_LOAD_DATA_LAZY_LOAD_DATA = "lazy_load_data"
    DICT_LOAD_DATA_ITER_TRAIN_DATA_LABELS = "iter_train_data_labels"
    DICT_LOAD_DATA_TOTAL_TRAIN_BATCH = "total_train_batch"
    DICT_LOAD_DATA_TRAIN_DATA = "train_data"
    DICT_LOAD_DATA_TEST_DATA = "test_data"
    DICT_LOAD_DATA_VALIDATION_DATA = "validation_data"
    DICT_LOAD_DATA_NUM_TRAIN_DATA = "num_train_data"
    DICT_LOAD_DATA_NUM_TEST_DATA = "num_test_data"
    DICT_LOAD_DATA_NUM_VALIDATION_DATA = "num_validation_data"
    DICT_LOAD_DATA_SHAPE_CHANNEL_HEIGHT_WIDTH = "shape_channel_height_width"



